'''
  Folding@Home Client Control (FAHControl)
  Copyright (C) 2010-2011 Stanford University

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import socket
import select
import time

debug = False

class ClientCom:
    def __init__(self, shutdown, address, port, password):
        self.shutdown = shutdown
        self.address = address
        self.port = int(port)
        self.password = password

        self.socket = None


    def is_connected(self):
        return self.socket is not None


    def close(self):
        if self.socket is not None:
            try:
                self.socket.shutdown(socket.SHUT_RDWR)
                self.socket.close()
            except: pass
            self.socket = None


    def connect(self):
        self.close()

        reason = 'connect'

        try:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.socket.connect((self.address, self.port))
            self.socket.setblocking(0)

            reason = 'flush'
            self.receive_result() # Flush to first command prompt

            reason = 'command'
            response = self.send_command('auth "%s"' % self.password)

            if response.strip() != 'OK':
                reason = 'access'
                raise Exception, 'Access denied'

            return 'ok'

        except Exception, e:
            print 'Connection to %s:%d failed: %s' % (
                self.address, self.port, e)
            self.close()

            return reason


    def flush(self):
        try:
            while not self.shutdown.isSet() and self.socket.recv(4096):
                continue
        except: pass


    def receive(self):
        try:
            buffer = self.socket.recv(4096)
            if buffer == '': raise Exception, 'Connection closed'

            return buffer

        except socket.error, (errno, msg):
            # Error codes for nothing to read
            if errno not in [errno.EAGAIN, errno.EWOULDBLOCK]: raise

        return ''
        

    def receive_block(self, timeout = 15):
        start = time.time()

        data = ''
        count = None
        more = False
        while True:
            buffer = self.receive()

            if buffer:
                data += buffer

                # Read count
                if count is None:
                    eol = data.find('\n')
                    if eol != -1:
                        line = data[:eol].strip()
                        if line:
                            if line[0] == '+':
                                more = True
                                line = line[1:]

                            data = data[eol + 1:]
                            count = int(line) - len(data)
                else:
                    count -= len(buffer)

            if count is not None and count <= -2: break
            if start + timeout < time.time():
                raise Exception, 'Receive timed out'

            self.shutdown.wait(0.1)
            if self.shutdown.isSet(): raise Exception, 'Shutdown'

        if count is not None and data[count:count + 2] == '\n>':
            data = data[:count]
            return more, data

        return False, ''


    def receive_result(self, timeout = 15):
        start = time.time()
        data = ''
        while True:
            buffer = self.receive()

            if buffer:
                found_end = (data[-1:] + buffer).find('\n>') != -1
                data += buffer
                if found_end or data[0] == '>': break

            if start + timeout < time.time():
                raise Exception, 'Command timed out'

            self.shutdown.wait(0.1)
            if self.shutdown.isSet(): raise Exception, 'Shutdown'

        if data[0] == '>': return ''
        return data[0:data.find('\n>')]


    def send_command(self, cmd, timeout = 15):
        if debug: print 'Sending: %s' % cmd

        self.flush()
        self.socket.send(cmd + '\n')

        return self.receive_result(timeout)


    def send(self, data):
        return self.socket.send(data)
